/*
 * table.c - table (listview report) dialog control DLL for mIRC
 * Copyright (c) 2006-2008, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <shellapi.h>

#define NEXT '\t'

#define ITEMBASE 6000

#define MAGIC 0x2a9fc835

#define LWL_MAGIC 0
#define LWL_NEXT 4
#define LWL_HEADER 8
#define LWL_COLUMNS 12
#define LWL_EVENT 16
#define LWL_MWND 20
#define LWL_SORT 24
#define LWL_LAST 28

#define WM_MCOMMAND (WM_USER + 200)
#define WM_MEVALUATE (WM_USER + 201)
#define WM_LVEVENT (WM_USER + 1000)

#define LOCALE_ENGLISH \
  MAKELCID(MAKELANGID(LANG_ENGLISH,SUBLANG_DEFAULT),SORT_DEFAULT)

#define malloc(s) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, s)
#define free(p) HeapFree(GetProcessHeap(), 0, p)

#define abs(n) (((n)<0)?(-(n)):(n))

static HRESULT (WINAPI *SetWindowTheme)(HWND hwnd, LPCWSTR pszSubAppName,
 LPCWSTR pszSubIdList);
static LRESULT CALLBACK BoxWndProc(HWND hwnd, UINT uMsg, WPARAM wParam,
 LPARAM lParam);

enum {
  OP_NONE,
  OP_ADD,
  OP_INSERT,
  OP_DELETE,
  OP_OVERWRITE,
  OP_SORT,
  OP_IMAGE,
  OP_ON,
  OP_OFF
};

enum {
  SORT_STRING,
  SORT_ISTRING,
  SORT_NUMERIC,
  SORT_FLAGS,
  SORT_BITFLAGS,
  SORT_ALIAS
};

typedef struct {
  int image;
  char *text;
} subitem_t;

typedef struct {
  int num;
  int size;
  subitem_t sub[1];
} item_t;

typedef struct {
  int column;
  int type;
  int invert;
  HWND hwnd;
  char *alias;
  unsigned long array[256];
  unsigned long defval;
} sort_t;

typedef struct {
  DWORD  mVersion;
  HWND   mHwnd;
  BOOL   mKeep;
} LOADINFO;

static unsigned char base64[64] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static HWND root = NULL;
static ATOM atom = NULL;

static WNDCLASSEX wcx = {
    sizeof(WNDCLASSEX),
    CS_PARENTDC,
    BoxWndProc,
    0,
    LWL_LAST,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    "ListViewBox",
    NULL
  };

void __stdcall LoadDll(LOADINFO *li) {
  wcx.hInstance = (HINSTANCE)GetWindowLong(li->mHwnd, GWL_HINSTANCE);

  atom = RegisterClassEx(&wcx);
}

int __stdcall UnloadDll(int timeout) {
  if (timeout == 1 && root != NULL) return 0;

  while (root != NULL)
   DestroyWindow(root);

  if (atom) UnregisterClass((LPCTSTR)MAKELONG(atom, 0), wcx.hInstance);

  return 1;
}

static BOOL send_mirc(HWND hwnd, int msg, char *data) {
  HANDLE hmap, hmutex;
  char *ptr;
  int res = FALSE;

  // create the filemapping object
  if ((hmap = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE,
   0, 1024, "mIRC")) == NULL) return FALSE;

  // at this point we SHOULD check GetLastError() to see if it's set to
  // ERROR_ALREADY_EXISTS; that's how mIRC prescribes that it must be done,
  // and if all DLL coders did this, we would not need a mutex at all. NONE
  // of them do, so I am forced to follow suit to retain compatibility. bah!

  // map the shared memory onto our address space
  if ((ptr = (char *)MapViewOfFile(hmap, FILE_MAP_ALL_ACCESS, 0, 0, 0)) ==
   NULL) {
    CloseHandle(hmap);

    return FALSE;
  }

  // the creation and mapping of the virtual file done so far, can not lead to
  // synchronization problems on itself; we only need to protect actual access

  // so, create or open a mutex that will protect the shared memory access
  if ((hmutex = CreateMutex(NULL, FALSE, "mIRC_SendMessage")) != NULL) {

    // wait for this mutex to be released - if it is currently locked, wait!
    WaitForSingleObject(hmutex, INFINITE);

    // now that we hold the mutex, we can access the shared memory safely
    lstrcpyn(ptr, data, 1024);

    // and tell mIRC to fulfill our request
    if (SendMessage(hwnd, msg, 0, NULL)) {
      res = TRUE;

      // and eventually, copy back the results
      if (msg == WM_MEVALUATE) lstrcpy(data, ptr);
    }

    // release and destroy the mutex
    ReleaseMutex(hmutex);

    CloseHandle(hmutex);
  }

  // unmap and close the filemapping
  UnmapViewOfFile(ptr);

  CloseHandle(hmap);

  return res;
}

static void notheme(HWND hwnd) {
  HINSTANCE hLibrary = LoadLibrary("uxtheme.dll");

  if (hLibrary != NULL) {
    SetWindowTheme = (HRESULT (WINAPI *)(HWND, LPCWSTR, LPCWSTR))
     GetProcAddress(hLibrary, "SetWindowTheme");

    if (SetWindowTheme != NULL)
     SetWindowTheme(hwnd, L"", L"");

    FreeLibrary(hLibrary);
  }
}

char *getnum(char *string, int *num) {
  int value = 0, filled = 0, neg = 0;
  char *ptr = string;

  while (*ptr == ' ') ptr++;

  if (*ptr == '-') {
    neg = 1;
    ptr++;
  }

  for ( ; *ptr >= '0' && *ptr <= '9'; filled = 1)
   value = value * 10 + *ptr++ - '0';

  if (!filled) return NULL;

  if (*ptr == ' ') ptr++;
  else if (*ptr) return NULL;

  if (neg) value = -value;

  *num = value;
  return ptr;
}

static char *getval(char *ptr, int *val) {
  char *next = getnum(ptr, val);

  if (next != NULL) return next;

  *val = -1;
  return ptr;
}

static char *gettok(char *ptr) {
  while (*ptr && *ptr != NEXT) ptr++;

  if (*ptr) *ptr++ = 0;

  return ptr;
}

int __stdcall create(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show,
 BOOL nopause) {
  HWND hwndDialog, hwndBox, hwndView;
  int style = LVS_REPORT, exstyle = 0;
  int item, prev = 0, x, y, w, h, theme = 1;
  char *ptr, *cb;

  // parse the options
  if ((ptr = getnum(data, &item)) == NULL) return 1;

  hwndDialog = (HWND)item;

  if ((ptr = getnum(ptr, &item)) == NULL) return 1;

  if (!IsWindow(hwndDialog) || GetDlgItem(hwndDialog, item) != NULL) return 1;

  if (*ptr == '+') {
    for (ptr++; *ptr && *ptr != ' '; ptr++) {
      switch (*ptr) {
        case 'c' : exstyle |= LVS_EX_CHECKBOXES; break;
        case 'f' : exstyle |= LVS_EX_FLATSB; break;
        case 'g' : exstyle |= LVS_EX_GRIDLINES; break;
        case 'h' : exstyle |= LVS_EX_HEADERDRAGDROP; break;
        case 'i' : exstyle |= LVS_EX_SUBITEMIMAGES; break;
        case 'l' : exstyle |= LVS_EX_LABELTIP; break;
        case 'n' : style |= LVS_NOCOLUMNHEADER; break;
        case 'o' : style |= LVS_NOSORTHEADER; break;
        case 'r' : exstyle |= LVS_EX_FULLROWSELECT; break;
        case 's' : style |= LVS_SINGLESEL; break;
        case 't' : theme = 0; break;
        case 'w' : style |= LVS_SHOWSELALWAYS; break;
      }
    }

    if (!*ptr++) return 1;
  }

  if ((ptr = getnum(ptr, &x)) == NULL) return 1;
  if ((ptr = getnum(ptr, &y)) == NULL) return 1;
  if ((ptr = getnum(ptr, &w)) == NULL) return 1;
  if ((ptr = getnum(ptr, &h)) == NULL) return 1;

  if ((cb = getnum(ptr, &prev)) == NULL) cb = ptr;

  InitCommonControls();

  // create the container for the listview
  hwndBox = CreateWindowEx(WS_EX_CONTROLPARENT, "ListViewBox", cb, WS_CHILD,
   x, y, w, h, hwndDialog, (HMENU)(ITEMBASE + item), wcx.hInstance, NULL);

  if (hwndBox == NULL) return 1;

  if (prev) SetWindowPos(hwndBox, GetDlgItem(hwndDialog, ITEMBASE + prev),
   0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE | SWP_NOOWNERZORDER);

  SetWindowLong(hwndBox, LWL_MAGIC, MAGIC);
  SetWindowLong(hwndBox, LWL_NEXT, (DWORD)root);
  SetWindowLong(hwndBox, LWL_HEADER, 0);
  SetWindowLong(hwndBox, LWL_COLUMNS, 0);
  SetWindowLong(hwndBox, LWL_EVENT, 0);
  SetWindowLong(hwndBox, LWL_MWND, (DWORD)mWnd);

  // now create the listview itself
  hwndView = CreateWindowEx(WS_EX_CLIENTEDGE, WC_LISTVIEW, "",
   WS_CHILD | WS_VISIBLE | WS_TABSTOP | style, 0, 0, w, h, hwndBox, (HMENU)1,
   wcx.hInstance, NULL);

  if (hwndView == NULL) {
    DestroyWindow(hwndBox);
    return 1;
  }

  (void)ListView_SetExtendedListViewStyle(hwndView, exstyle);

  if (!theme) notheme(hwndView);

  // and show the windows
  ShowWindow(hwndBox, SW_SHOWNA);

  // finally link in the window into the linked list
  root = hwndBox;

  return 1;
}

static char *getctrl(char *ptr, HWND *hptr, HWND *hbox) {
  HWND hwnd;
  int item;

  // get the dialog HWND
  if ((ptr = getnum(ptr, &item)) == NULL) return NULL;

  hwnd = (HWND)item;

  // get the dialog item number
  if ((ptr = getnum(ptr, &item)) == NULL) return NULL;

  // get the box HWND from those two
  hwnd = GetDlgItem(hwnd, ITEMBASE + item);

  if (hwnd == NULL || GetWindowLong(hwnd, LWL_MAGIC) != MAGIC) return NULL;

  if (hbox) *hbox = hwnd;

  // and get the child HWND from the box HWND - it's its only child
  *hptr = GetWindow(hwnd, GW_CHILD);

  return ptr;
}

static char *getcolumn(LVCOLUMN *lvc, char *ptr) {
  char *next, *text, *flags = ptr;

  next = gettok(ptr);

  while (*ptr == ' ') ptr++;

  while (*ptr && *ptr != ' ') ptr++;

  if (*ptr) *ptr++ = 0;

  text = getnum(ptr, &lvc->cx);

  lvc->mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;

  if (text != NULL) {
    lvc->mask |= LVCF_WIDTH;
    lvc->pszText = text;
  }
  else lvc->pszText = ptr;

  lvc->fmt = 0;
  while (*flags) {
    switch (*flags++) {
      case 'c' : lvc->fmt = LVCFMT_CENTER; break;
      case 'r' : lvc->fmt = LVCFMT_RIGHT; break;
    }
  }

  return next;
}

int __stdcall header(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show,
 BOOL nopause) {
  LVCOLUMN lvc;
  char *ptr, *copy;
  HWND hwnd, hbox;
  int column;

  if ((ptr = getctrl(data, &hwnd, &hbox)) == NULL) return 1;

  // remove all the old columns (if any)
  for (column = GetWindowLong(hbox, LWL_COLUMNS); column > 0; column--)
   ListView_DeleteColumn(hwnd, column - 1);

  // free the copy of the old header definition (if any)
  copy = (char *)GetWindowLong(hbox, LWL_HEADER);
  if (copy != NULL) free(copy);

  // store a copy of the new header definition
  copy = malloc(lstrlen(ptr) + 1);
  lstrcpy(copy, ptr);
  SetWindowLong(hbox, LWL_HEADER, (DWORD)copy);

  lvc.mask = 0;

  // insert a dummy column, because the first one is always left-aligned
  ListView_InsertColumn(hwnd, column++, &lvc);

  // insert all other columns
  while (*ptr) {
    ptr = getcolumn(&lvc, ptr);

    ListView_InsertColumn(hwnd, column++, &lvc);
  }

  // remove the dummy column
  ListView_DeleteColumn(hwnd, 0);

  // store the new number of columns
  SetWindowLong(hbox, LWL_COLUMNS, column - 1);

  return 1;
}

int __stdcall widths(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show,
 BOOL nopause) {
  int width, column, total;
  HWND hwnd, hbox;
  char *ptr;

  if ((ptr = getctrl(data, &hwnd, &hbox)) == NULL) return 1;

  total = GetWindowLong(hbox, LWL_COLUMNS);

  for (column = 0; column < total && ptr && *ptr; column++) {
    while (*ptr == ' ') ptr++;

    if ((*ptr == 'a' || *ptr == 'h') && (!ptr[1] || ptr[1] == ' '))
     width = (*(ptr++) == 'a') ? LVSCW_AUTOSIZE : LVSCW_AUTOSIZE_USEHEADER;
    else if ((ptr = getnum(ptr, &width)) == NULL) break;

    ListView_SetColumnWidth(hwnd, column, width);
  }

  return 1;
}

static int setimage(HWND hwnd, char *ptr) {
  HIMAGELIST list;
  HICON icon;
  int index;

  if ((ptr = getnum(ptr, &index)) == NULL || index < 0) return 1;

  // first check how many icons the file actually contains
  icon = ExtractIcon(wcx.hInstance, ptr, (UINT)-1);

  if ((int)icon <= 0 || (int)icon <= index) return 1;

  // get the small icons image list - if there wasn't one, create one
  list = ListView_GetImageList(hwnd, LVSIL_SMALL);

  if (list == NULL) {
    list = ImageList_Create(GetSystemMetrics(SM_CXSMICON),
     GetSystemMetrics(SM_CYSMICON), ILC_COLOR32 | ILC_MASK, 8, 8);

    if (list == NULL) return 1;

    (void)ListView_SetImageList(hwnd, list, LVSIL_SMALL);
  }

  // now extract the actual icon, and add it
  icon = ExtractIcon(wcx.hInstance, ptr, index);

  if (icon == NULL || (int)icon == 1) return 1;

  ImageList_AddIcon(list, icon);

  DestroyIcon(icon);

  return 1;
}

static item_t *create_item(char *str) {
  int i, val, len, num = 0;
  char *ptr, *next;
  item_t *item;

  len = lstrlen(str) + 1;
  ptr = str;

  // find out how many subitems there are
  // also, make sure the input is actually valid
  while (*ptr) {
    ptr = gettok(ptr);

    num++;
  }

  // there must be at least one subitem, namely the primary item
  if (num == 0) return NULL;

  // allocate memory for the item and its subitems (including strings)
  item = (item_t *)
   malloc(sizeof(item_t) + sizeof(subitem_t) * (num - 1) + len);
  if (item == NULL) return NULL;

  // initialize the number of subitems, and string of the string data
  item->num = num;
  item->size = len;

  // store all subitems, including the full text string and
  // block-internal pointers into this string (for efficient lookups)
  ptr = (char *)(&item[1]) + sizeof(subitem_t) * (num - 1);

  CopyMemory(ptr, str, len);

  for (i = 0; i < num; i++) {
    for (next = ptr; *next == ' '; next++);

    while (*next && *next != ' ') next++;

    if (*next) next++;

    getval(ptr, &val);

    item->sub[i].image = val;
    item->sub[i].text = next;

    ptr += lstrlen(ptr) + 1;
  }

  return item;
}

static void free_item(item_t *item) {
  // we use one memory chunk for the whole item structure, so this is simple
  free(item);
}

static void fill_item(LVITEM *lvi) {
  item_t *item;

  // get our item_t structure from the LVITEM structure
  if ((item = (item_t *)lvi->lParam) == NULL ||
   lvi->iSubItem >= item->num) return;

  // fill other parts of the LVITEM structure as requested
  if (lvi->mask & LVIF_TEXT) lvi->pszText = item->sub[lvi->iSubItem].text;
  if (lvi->mask & LVIF_IMAGE) lvi->iImage = item->sub[lvi->iSubItem].image;
}

static char *encode(unsigned char *in, unsigned char *out) {
  unsigned int val = 0, bits = 24;

  // standard base64 encoding
  while (*in) {
    bits -= 8;
    val |= (*in << bits);

    if (bits == 0) {
      *out++ = base64[(val >> 18) & 0x3f];
      *out++ = base64[(val >> 12) & 0x3f];
      *out++ = base64[(val >> 6) & 0x3f];
      *out++ = base64[val & 0x3f];
      val = 0;
      bits = 24;
    }

    in++;
  }

  if (bits < 24) {
    *out++ = base64[(val >> 18) & 0x3f];
    *out++ = base64[(val >> 12) & 0x3f];
    *out++ = (bits < 12) ? base64[(val >> 6) & 0x3f] : '=';
    *out++ = '=';
  }

  return out;
}

static int compare_alias(sort_t *sort, char *str1, char *str2) {
  char buf[4096], *ptr;
  int val;

  // construct an identifier call that will do the comparison; for this it
  // needs the two input strings. we have to use $decode(<encoded string>)
  // to prevent evaluation of the strings' contents, as well as any problems
  // due to commas, parentheses, brackets etc..
  ptr = buf + wsprintf(buf, "$%s($decode(", sort->alias);

  ptr = encode(str1, ptr);

  ptr += wsprintf(ptr, ",m),$decode(");

  ptr = encode(str2, ptr);

  lstrcpy(ptr, ",m))");

  // we can use over 3500 bytes of the buffer this way - cut it off
  buf[900] = 0;

  // now let mIRC evaluate the identifier
  if (send_mirc(sort->hwnd, WM_MEVALUATE, buf) && getnum(buf, &val))
   return val;

  return 0;
}

static int compare_flags(sort_t *sort, unsigned char *str1,
 unsigned char *str2) {
  unsigned long val1, val2;

  for (val1 = 0; *str1; str1++)
   if (val1 < sort->array[*str1]) val1 = sort->array[*str1];

  if (val1 == 0) val1 = sort->defval;

  for (val2 = 0; *str2; str2++)
   if (val2 < sort->array[*str2]) val2 = sort->array[*str2];

  if (val2 == 0) val2 = sort->defval;

  return val2 - val1;
}

static int compare_bitflags(sort_t *sort, unsigned char *str1,
 unsigned char *str2) {
  unsigned long val1, val2;

  for (val1 = 0; *str1; str1++)
   val1 |= sort->array[*str1];

  if (val1 == 0) val1 = sort->defval;

  for (val2 = 0; *str2; str2++)
   val2 |= sort->array[*str2];

  if (val2 == 0) val2 = sort->defval;

  return val2 - val1;
}

static int compare_numeric(char *str1, char *str2) {
  int val1, val2;

  str1 = getnum(str1, &val1);
  str2 = getnum(str2, &val2);

  return val1 - val2;
}

static int compare_items(sort_t *sort, item_t *item1, item_t *item2) {
  char *str1, *str2;
  int flag = SORT_STRINGSORT;

  // put the items without data for this column, at the bottom
  if (item1->num <= sort->column)
   return (item2->num <= sort->column) ? 0 : 1;
  else if (item2->num <= sort->column) return -1;

  str1 = item1->sub[sort->column].text;
  str2 = item2->sub[sort->column].text;

  switch (sort->type) {
    case SORT_ISTRING : flag |= NORM_IGNORECASE; /* fall-through */
    case SORT_STRING :
      return CompareString(LOCALE_ENGLISH, flag, str1, -1, str2, -1) - 2;
    case SORT_NUMERIC : return compare_numeric(str1, str2);
    case SORT_FLAGS : return compare_flags(sort, str1, str2);
    case SORT_BITFLAGS : return compare_bitflags(sort, str1, str2);
    case SORT_ALIAS : return compare_alias(sort, str1, str2);
  }

  return 0;
}

static int CALLBACK compare(LPARAM lp1, LPARAM lp2, LPARAM lps) {
  sort_t *sort = (sort_t *)lps;

  int res = compare_items(sort, (item_t *)lp1, (item_t *)lp2);

  return (sort->invert) ? -res : res;
}

static void clear(HWND hwnd) {
  LVITEM lvi;
  int num;

  lvi.mask = LVIF_PARAM;
  lvi.iSubItem = 0;

  for (num = ListView_GetItemCount(hwnd) - 1; num >= 0; num--) {
    lvi.iItem = num;

    if (ListView_GetItem(hwnd, &lvi))
     free_item((item_t *)lvi.lParam);
  }

  ListView_DeleteAllItems(hwnd);
}

static BOOL do_op(HWND hwnd, HWND hbox, int op, int num, char *ptr) {
  LVITEM lvi;
  item_t *item;
  sort_t sort;

  switch (op) {
    case OP_ADD :
    case OP_INSERT :
      if ((item = create_item(ptr)) == NULL) return FALSE;

      lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
      lvi.iItem = num;
      lvi.iSubItem = 0;
      lvi.pszText = LPSTR_TEXTCALLBACK;
      lvi.iImage = I_IMAGECALLBACK;
      lvi.lParam = (LPARAM)item;

      if (ListView_InsertItem(hwnd, &lvi) < 0) {
        free_item(item);

        return FALSE;
      }

      break;
    case OP_OVERWRITE :
      lvi.mask = LVIF_PARAM;
      lvi.iItem = num;
      lvi.iSubItem = 0;

      if (!ListView_GetItem(hwnd, &lvi)) return FALSE;

      if ((item = create_item(ptr)) == NULL) return FALSE;

      free_item((item_t *)lvi.lParam);

      lvi.lParam = (LPARAM)item;

      ListView_SetItem(hwnd, &lvi);

      break;
    case OP_DELETE :
      lvi.mask = LVIF_PARAM;
      lvi.iItem = num;
      lvi.iSubItem = 0;

      if (!ListView_GetItem(hwnd, &lvi)) return FALSE;

      ListView_DeleteItem(hwnd, num);

      free_item((item_t *)lvi.lParam);

      break;
    case OP_SORT :
      if (num < 0) return FALSE;

      sort.column = num;
      sort.type = SORT_STRING;
      sort.invert = 0;
      sort.hwnd = (HWND)GetWindowLong(hbox, LWL_MWND);

      while (*ptr && *ptr != ' ') {
        switch (*ptr++) {
          case 'd' : // 'd' (descending) legacy option, now renamed to 'r'
          case 'r' : sort.invert = 1; break;
          case 'n' : sort.type = SORT_NUMERIC; break;
          case 'i' : sort.type = SORT_ISTRING; break;
          case 'f' : sort.type = SORT_FLAGS; break;
          case 'b' : sort.type = SORT_BITFLAGS; break;
          case 'a' : sort.type = SORT_ALIAS; break;
        }
      }

      while (*ptr == ' ') ptr++;

      if (sort.type == SORT_ALIAS) sort.alias = ptr;
      else if (sort.type == SORT_FLAGS || sort.type == SORT_BITFLAGS) {
        unsigned char *p = ptr;
        unsigned long c;

        ZeroMemory(sort.array, sizeof(sort.array));
        sort.defval = 0;

        if (sort.type == SORT_FLAGS) {
          for (c = 255; *p; p++, c--)
           if (*p == NEXT) sort.defval = c;
           else sort.array[*p] = c;
        }
        else {
          for (c = 1 << 30; *p; p++, c >>= 1)
           if (*p == NEXT) sort.defval = c;
           else sort.array[*p] = c;
        }

      }

      ListView_SortItems(hwnd, compare, (LPARAM)&sort);

      num++;
      SetWindowLong(hbox, LWL_SORT, (sort.invert) ? -num : num);

      break;
  }

  return TRUE;
}

int __stdcall set(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show,
 BOOL nopause) {
  char *ptr;
  HWND hwnd, hbox;
  int num, op = OP_NONE, select = OP_NONE, check = OP_NONE;

  if ((ptr = getctrl(data, &hwnd, &hbox)) == NULL) return 1;

  if (*ptr != '-') return 1;

  for (ptr++; *ptr && *ptr != ' '; ptr++) {
    switch (*ptr) {
      case 'f' : SetFocus(hwnd); break;
      case 'e' : EnableWindow(hwnd, TRUE); break;
      case 'b' : EnableWindow(hwnd, FALSE); break;
      case 'v' : ShowWindow(hbox, SW_SHOWNA); break;
      case 'h' : ShowWindow(hbox, SW_HIDE); break;

      case 'r' : clear(hwnd); break;

      case 'a' : op = OP_ADD; break;
      case 'd' : op = OP_DELETE; break;
      case 'i' : op = OP_INSERT; break;
      case 'o' : op = OP_OVERWRITE; break;
      case 's' : op = OP_SORT; break;

      case 'c' : select = OP_ON; break;
      case 'u' : select = OP_OFF; break;
      case 'x' : check = OP_ON; break;
      case 'l' : check = OP_OFF; break;

      case 'g' : op = OP_IMAGE; break;
    }
  }

  if (!*ptr || (op == OP_NONE && select == OP_NONE && check == OP_NONE))
   return 1;
  ptr++;

  if (op == OP_IMAGE) return setimage(hwnd, ptr);

  if (op != OP_ADD) {
    ptr = getnum(ptr, &num);

    if (ptr == NULL) return 1;

    num--;
  }
  else num = ListView_GetItemCount(hwnd);

  if (!do_op(hwnd, hbox, op, num, ptr)) return 1;

  if (op != OP_DELETE && op != OP_SORT) {
    if (select != OP_NONE) ListView_SetItemState(hwnd, num,
      (select == OP_ON) ? LVIS_SELECTED : 0, LVIS_SELECTED);
    if (check != OP_NONE) ListView_SetCheckState(hwnd, num, (check == OP_ON));
  }

  return 1;
}

int __stdcall get(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show,
 BOOL nopause) {
  char *ptr;
  HWND hwnd, hbox;
  LVITEM lvi;
  item_t *item;
  int i, sel, num = 0;

  if ((ptr = getctrl(data, &hwnd, &hbox)) == NULL) return 1;

  ptr = getval(ptr, &num);

  // [item] sel, seltext
  sel = lstrcmpi(ptr, "sel");
  if (!sel || !lstrcmpi(ptr, "seltext")) {
    if (!num) {
      wsprintf(data, "%d", ListView_GetSelectedCount(hwnd));

      return 3;
    }

    if ((i = num) < 0) return 1;

    for (num = -1; i > 0; i--) {
      num = ListView_GetNextItem(hwnd, num, LVNI_SELECTED);

      if (num < 0) return 1;
    }

    num++;

    if (!sel) {
      wsprintf(data, "%d", num);

      return 3;
    }

    /* fall-through for seltext */
  }

  // header
  if (!lstrcmpi(ptr, "header")) {
    char *header;

    header = (char *)GetWindowLong(hbox, LWL_HEADER);

    if (header != NULL) lstrcpy(data, header);
    else *data = 0;

    return 3;
  }

  // sort
  if (!lstrcmpi(ptr, "sort")) {
    wsprintf(data, "%d", GetWindowLong(hbox, LWL_SORT));

    return 3;
  }

  // (item count)
  if (!num) {
    wsprintf(data, "%d", ListView_GetItemCount(hwnd));

    return 3;
  }

  lvi.mask = LVIF_STATE | LVIF_PARAM;
  lvi.iItem = num - 1;
  lvi.iSubItem = 0;
  lvi.stateMask = LVIS_SELECTED | LVIS_FOCUSED;

  if (!ListView_GetItem(hwnd, &lvi)) return 1;

  // <item> check
  if (!lstrcmpi(ptr, "check")) {
    wsprintf(data, "%d", ListView_GetCheckState(hwnd, num - 1));

    return 3;
  }

  // <item> state
  if (!lstrcmpi(ptr, "state")) {
    int mask = 0;

    if (lvi.state & LVIS_SELECTED) mask |= 1;
    if (lvi.state & LVIS_FOCUSED) mask |= 2;

    wsprintf(data, "%d", mask);

    return 3;
  }

  // <item>
  item = (item_t *)lvi.lParam;

  CopyMemory(data, (char *)(&item[1]) + sizeof(subitem_t) * (item->num - 1),
   item->size);

  for (i = 0; i < item->size - 1; i++)
   if (!data[i]) data[i] = NEXT;

  return 3;
}

static void notify(HWND hwnd, char *name, int num) {
  // a "num" value of -2 means "no number"; -1 and up are actual numbers

  // if there was still a "sclick" event pending, send it first (recursively)
  if (GetWindowLong(hwnd, LWL_EVENT)) {
    SetWindowLong(hwnd, LWL_EVENT, 0);

    notify(hwnd, "sclick", -2);
  }

  // send the notification command to the main mIRC window
  if (name != NULL) {
    char buf[1024], *ptr;

    buf[0] = '/';

    ptr = buf + 1;
    ptr += GetWindowText(hwnd, ptr, sizeof(buf) - lstrlen(name) - 17);

    // if there actually is a callback function, send it to mIRC
    if (ptr != buf + 1) {
      if (num == -2) wsprintf(ptr, " %s", name);
      else wsprintf(ptr, " %s %d", name, num);

      send_mirc((HWND)GetWindowLong(hwnd, LWL_MWND), WM_MCOMMAND, buf);
    }
  }
}

static void notify_click(HWND hwnd, char *name, LPARAM lParam) {
  NMITEMACTIVATE *nmia = (NMITEMACTIVATE *)lParam;

  // the initial version of NMITEMACTIVATE did not have the "iItem" element;
  // this is a really ugly hack to still support it for later versions
  notify(hwnd, name, IsBadReadPtr(&nmia->iItem, sizeof(nmia->iItem)) ? -2 :
   nmia->iItem + 1);
}

static LRESULT CALLBACK BoxWndProc(HWND hwnd, UINT uMsg, WPARAM wParam,
 LPARAM lParam) {
  if (uMsg == WM_NOTIFY) {
    LPNMHDR nm = (LPNMHDR)lParam;

    switch (nm->code) {
      case LVN_GETDISPINFO :
        fill_item(&(((LV_DISPINFO *)lParam)->item));
        break;
      case LVN_COLUMNCLICK :
        notify(hwnd, "hclick", ((NMLISTVIEW *)lParam)->iSubItem + 1);

        break;
      case LVN_ITEMCHANGED :
        // schedule a "sclick" event; we do it this way to prevent a flood
        // of notifications because of all the single ITEMCHANGED messages
        if ((((NMLISTVIEW *)lParam)->uChanged & LVIF_STATE) &&
         !GetWindowLong(hwnd, LWL_EVENT)) {
          SetWindowLong(hwnd, LWL_EVENT, 1);

          PostMessage(hwnd, WM_LVEVENT, 0, 0);
        }

        break;
      case NM_RETURN :
      case NM_DBLCLK :
        notify_click(hwnd, "dclick", lParam);

        break;
      case NM_RCLICK :
      case NM_RDBLCLK :
        if (wParam > 0) notify_click(hwnd, "rclick", lParam);
        else notify(hwnd, "rclick", -1);

        break;
    }
  }
  else if (uMsg == WM_DESTROY) {
    HWND nptr;

    // free the stored header definition (if any)
    char *header = (char *)GetWindowLong(hwnd, LWL_HEADER);
    if (header != NULL) free(header);

    // free all the item_t structures associated with items
    clear(GetWindow(hwnd, GW_CHILD));

    // unlink the window from the linked list

    nptr = (HWND)GetWindowLong(hwnd, LWL_NEXT);

    if (root == hwnd) root = nptr;
    else {
      HWND cur, next;

      for (cur = root; cur != NULL; cur = next) {
        if ((next = (HWND)GetWindowLong(cur, LWL_NEXT)) == hwnd) {
          SetWindowLong(cur, LWL_NEXT, (DWORD)nptr);
          break;
        }
      }
    }
  }
  else if (uMsg == WM_LVEVENT) {
    // send the pending "sclick" event
    notify(hwnd, NULL, 0);

    return 0;
  }

  return DefWindowProc(hwnd, uMsg, wParam, lParam);
}
